from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from itertools import chain
import math
import hashlib
import struct

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_group_metadata = KEYS[1];
local redis_key_group_names = KEYS[2];
local group_key = ARGV[1];
local group_name_with_key_new = ARGV[2];
local group_name_with_key_old = ARGV[3];
local stream_id = redis.call("XADD", redis_key_group_metadata, "*", unpack(ARGV, 4, table.maxn(ARGV)));
redis.call("ZREM", redis_key_group_names, group_name_with_key_old);
redis.call("ZADD", redis_key_group_names, 0, group_name_with_key_new);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.update_group(event.session, **event.data)

    async def update_group(self, session : EventSession, group_key : str = '', gid : str = '', **update_params):
        if not group_key:
            raise ValueError('group_key is required.')
        
        metadata, version = await self.helper.get_group_metadata_with_version(self.manager.redis, group_key)
        old = metadata.copy()
        
        if gid and gid != metadata.gid:
            raise ValueError('gid was expeccted to be [{}] but was [{}].'.format(metadata.gid, gid))

        metadata.update(update_params)

        if old == metadata:
            return version

        redis_key_group_metadata = self.helper.stream_key_for_group_metadata(group_key)
        
        if old.group_name == metadata.group_name:
            return await self.manager.redis.xadd(redis_key_group_metadata, **metadata)
        else:
            redis_key_group_names = self.helper.zset_key_for_group_names()
            ret = await session.redis.evalsha(
                Handler.SCRIPT
                , 2
                , redis_key_group_metadata, redis_key_group_names
                , metadata.group_key, '{}::{}'.format(metadata.group_name,metadata.group_key), '{}::{}'.format(old.group_name,metadata.group_key), *chain.from_iterable(metadata.items()))
            return metadata.group_key



from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from struct import pack
from itertools import chain
import math
import hashlib

from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_contents = KEYS[1];
local redis_key_contents_object = KEYS[2];
local redis_key_contents_keys = KEYS[3];
local content_key = ARGV[1];
local content_order = ARGV[2];
local content = ARGV[3];
if content ~= "" then
  redis.call("SET", redis_key_contents_object, content);
end
local stream_id = redis.call("XADD", redis_key_contents, "*", unpack(ARGV, 4, table.maxn(ARGV)));
redis.call("ZADD", redis_key_contents_keys, content_order, content_key);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.add_content(event.session, event.data[0], event.data[1], **event.data[2])

    
    async def add_content(
            self
            , session : EventSession
            , group_key : str
            , content : bytes
            , content_key : str = ''
#            , content_name : str = ''
            , content_type : str = ''
            , encoding : str =''
            , last_modified : str = ''
            , order : int = ''
            , **other_params):
        if not group_key:
            raise ValueError('group_key must be set')
        if not content:
            raise ValueError('content must be set')

        group = await self.helper.get_group_metadata(self.manager.redis, group_key)
        metadata = await self.helper.get_content_metadata(other_params.copy()
                                                          , content_key
                                                          , content_type
                                                          , encoding
                                                          , last_modified)

        if isinstance(content, str):
            metadata.content_type = metadata.content_type if metadata.content_type else 'text/plain'
            encoding = encoding if encoding else 'UTF-8'
            metadata.encoding = encoding
            content = content.encode(encoding)
        elif isinstance(content, bytes):
            metadata.content_type = metadata.content_type if metadata.content_type else 'application/octet-stream'
            if encoding:
                metadata.encoding = encoding
        else:
            raise ValueError('content must be str or bytes but was {}'.format(type(content)))
        
        metadata.content_length = len(content)
        metadata.cid = hashlib.sha1(content).hexdigest()

        index = await session.redis.execute('INCR', self.helper.incr_key_for_content_id(group.gid))
        metadata.order = order if order else index * 10
        
        redis_key_contents_object = self.helper.obj_key_for_content(group.gid, metadata.cid)
        if int(await session.redis.execute('EXISTS', redis_key_contents_object)) > 0:
            content = ''
            
        if metadata.content_key:
            if int(await session.redis.execute('EXISTS', self.helper.stream_key_for_contents_metadata(group.gid, metadata.content_key))) > 0:
                raise KeyError('content_key:[{}/{}] already exits.'.format(group_key, content_key))
            else:
                pass
        else:
            metadata.content_key = hashlib.sha1(redis_key_contents_object.encode('UTF-8')).hexdigest()
        
        redis_key_contents_metadata = self.helper.stream_key_for_contents_metadata(group.gid, metadata.content_key)
        redis_key_contents_keys = self.helper.zset_key_for_content_keys(group.gid)
        
        ret = await session.redis.evalsha(
            Handler.SCRIPT
            , 3
            , redis_key_contents_metadata, redis_key_contents_object, redis_key_contents_keys
            , metadata.content_key, metadata.order, content, *chain.from_iterable(metadata.items()))
        return {'group_key': group.group_key, 'content_key': metadata.content_key}
            





import types

def nameddict(name, dict_keys: str):
    keys = []
    if isinstance(dict_keys, str):
        keys = [v for v in dict_keys.split(' ') if v]
    elif isinstance(dict_keys, (list, tuple)):
        keys = [v for v in [val.strip() for val in dict_keys] if v]
    else:
        ValueError('dict_keys must be str or list but was [{}]'.format(type(dict_keys)))
    
    def create_property(attr):
        def getter(self):
            try:
                return self[attr]
            except KeyError:
                return None

        def setter(self, value):
            self[attr] = value
        return property(getter, setter)
        
    def init(self, dict={}):
        NamedDict.__init__(self, dict)
    
    def setattribute(self, attr, val):
        if not hasattr(self, attr):
            raise AttributeError("'{}' object has no attribute '{}'".format(name, attr))
        NamedDict.__setattr__(self, attr, val)

    def update_ns(ns):
        for attr in keys:
            ns[attr] = create_property(attr)
        ns['KEYS'] = keys
        ns['__init__'] = init
        ns['__setattr__'] = setattribute

    return types.new_class(name, bases=(NamedDict,), exec_body=update_ns)
    

class NamedDict(dict):

    def __init__(self, dict = None):
        super().__init__(dict)

    def __str__(self):
        return '{}[{}]'.format(self.__class__.__name__, super().__str__())

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, super().__repr__())

    def copy(self):
        return self.__class__(super().copy())




from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from itertools import chain
import math
import hashlib
import struct

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_group_metadata = KEYS[1];
local redis_key_group_names = KEYS[2];
local group_key = ARGV[1];
local group_name_with_key = ARGV[2];
local stream_id = redis.call("XADD", redis_key_group_metadata, "*", unpack(ARGV, 3, table.maxn(ARGV)));
redis.call("ZADD", redis_key_group_names, 0, group_name_with_key);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.add_group(event.session, **event.data[0])

    async def add_group(self, session : EventSession, group_name : str = '', content_type : str = '', group_key : str = '', **other_params):
        metadata = self.helper.GroupMetadata(other_params.copy())
        metadata.group_name = group_name if group_name else ''
        metadata.content_type = content_type if content_type else 'application/octet-stream'

        if group_key:
            if self.helper.is_compatible_with_sha1_key(group_key):
                metadata.group_key = group_key
            else:
                metadata.group_key = hashlib.sha1(group_key.encode('UTF-8')).hexdigest()
                metadata.group_key_text = group_key
            if int(await session.redis.execute('EXISTS', self.helper.stream_key_for_group_metadata(metadata.group_key))) > 0:
                raise KeyError('group_key:[{}] already exits.'.format(group_key))
            metadata.gid = await session.redis.execute('INCR', self.helper.incr_key_for_group_id())
        else:
            metadata.gid = await session.redis.execute('INCR', self.helper.incr_key_for_group_id())
            metadata.group_key = group_key if group_key else hashlib.sha1(struct.pack('i', metadata.gid)+struct.pack('d', datetime.now().timestamp())).hexdigest()
        
        redis_key_group_metadata = self.helper.stream_key_for_group_metadata(metadata.group_key)
        redis_key_group_names = self.helper.zset_key_for_group_names()

        ret = await session.redis.evalsha(
            Handler.SCRIPT
            , 2
            , redis_key_group_metadata, redis_key_group_names
            , metadata.group_key, '{}::{}'.format(metadata.group_name,metadata.group_key), *chain.from_iterable(metadata.items()))
        return metadata.group_key
            





from ducts.spi import EventHandler, webapi

from aiohttp import web

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        handler_spec.set_description('イベント情報を返します。')
        handler_spec.set_as_responsive()
        return handler_spec

    async def handle(self, event):
        return await event.session.redis.execute(event.data[0], *event.data[1:])

    @webapi.add_route(path='/', method='GET')
    async def get(self, request):
        return web.json_response(self.manager.key_ids)

    @webapi.add_route(path='/{id}', method='GET')
    async def get_by_id(self, request):
        eid = int(request.match_info['id'])
        return web.json_response(self.manager.get_handler_for(eid)[1].SPEC._asdict())




from ducts.spi import EventHandler

from io import BytesIO
from datetime import datetime
from hashlib import md5, sha1, sha256
from struct import pack
from itertools import chain

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('open buffer')
        return handler_spec

    async def handle(self, event):
        return await self.open_buffer(event.session)

    async def open_buffer(self, session):
        redis_key_buffer_keys = self.helper.stream_key_for_object_buffer()
        kv = {'sid': session.session_id()}
        stream_id = await session.redis.execute('XADD', redis_key_buffer_keys, '*', *chain.from_iterable(kv.items()))
        buffer_id = md5(stream_id).hexdigest()
        redis_key_for_object_buffer = self.helper.obj_key_for_object_buffer(session.session_id(), buffer_id)
        await session.redis.execute('SET', redis_key_for_object_buffer, '')
        return buffer_id
        



from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from itertools import chain
import math
import hashlib
import struct

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_contents = KEYS[1];
local redis_key_contents_keys = KEYS[2];
local content_key = ARGV[1];
local content_order = ARGV[2];
local stream_id = redis.call("XADD", redis_key_contents, "*", unpack(ARGV, 3, table.maxn(ARGV)));
redis.call("ZADD", redis_key_contents_keys, content_order, content_key);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.revert_content(event.session, event.data[0], event.data[1], event.data[2])

    async def revert_content(
            self
            , session : EventSession
            , group_key : str
            , content_key : str
            , to_version : str):
        
        if not group_key:
            raise ValueError('group_key must be set')
        if not content_key:
            raise ValueError('content_key must be set')
        if not to_version:
            raise ValueError('to_version must be set')

        
        group, to_content, to_version = await self.helper.get_group_content_metadata_for(
            self.manager.redis, group_key, content_key, to_version)
        
        group, from_content, from_version = await self.helper.get_group_content_metadata_with_version(
            self.manager.redis, group_key, content_key)

        if to_content == from_content:
            return to_version

        if from_content.content_key != to_content.content_key:
            raise ValueError('content_key was expeccted to be [{}] but was [{}].'.format(to_content.content_key, from_content.content_key))

        if not to_content.revert_to:
            to_content.revert_to = to_version

        redis_key_contents_metadata = self.helper.stream_key_for_contents_metadata(group.gid, to_content.content_key)
        redis_key_contents_keys = self.helper.zset_key_for_content_keys(group.gid)
        
        ret = await session.redis.evalsha(
            Handler.SCRIPT
            , 2
            , redis_key_contents_metadata, redis_key_contents_keys
            , to_content.content_key, to_content.order, *chain.from_iterable(to_content.items()))
        return {'group_key': group.group_key, 'content_key': to_content.content_key, 'version': ret.decode('UTF-8')}




from ducts.spi import EventHandler

from io import BytesIO
from datetime import datetime
from hashlib import md5, sha1, sha256
from struct import pack
from itertools import chain

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('add buffer data')
        return handler_spec

    async def handle(self, event):
        return await self.add_to_buffer(event.session, event.data[0], event.data[1])

    async def add_to_buffer(self, session, buffer_id, buffer_data):
        if not buffer_id:
            raise ValueError('buffer_id cannot be null')
        if not buffer_data:
            raise ValueError('buffer data cannot be null')

        redis_key_for_object_buffer = self.helper.obj_key_for_object_buffer(session.session_id(), buffer_id)
        
        if not await session.redis.execute('EXISTS', redis_key_for_object_buffer):
            raise ValueError('buffer_id[{}] not found in session[{}]'.format(buffer_id, session.session_id()))

        return await session.redis.execute('APPEND', redis_key_for_object_buffer, buffer_data)



from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from itertools import chain
import math
import hashlib
import struct

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_group_metadata = KEYS[1];
local redis_key_group_names = KEYS[2];
local group_key = ARGV[1];
local group_name_with_key_new = ARGV[2];
local group_name_with_key_old = ARGV[3];
local stream_id = redis.call("XADD", redis_key_group_metadata, "*", unpack(ARGV, 4, table.maxn(ARGV)));
redis.call("ZREM", redis_key_group_names, group_name_with_key_old);
redis.call("ZADD", redis_key_group_names, 0, group_name_with_key_new);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.update_group(event.session, event.data)

    async def update_group(self, session : EventSession, group_key : str = ''):
        metadata = await self.helper.get_group_metadata(self.manager.redis, group_key)
        count = 0
        
        scan_key = self.helper.scan_key_blobs_gid(metadata.gid)
        async for key in session.redis.scan_for(scan_key):
            ret = await session.redis.execute('DEL', key)
            logger.info('DELETE|%s|%Ss', key, ret)
            count += 1
        redis_key_group_metadata = self.helper.stream_key_for_group_metadata(metadata.group_key)
        ret = await session.redis.execute('DEL', redis_key_group_metadata)
        count += 1
        
        return count




import hashlib
from datetime import datetime
from email.utils import parsedate_to_datetime
from mimetypes import guess_type

import asyncio

from ducts import utils
from ifconf import configure_module, config_callback

import logging
logger = logging.getLogger(__name__)
        
@config_callback
def config(loader):
    pass

conf = configure_module(config)

GroupMetadata = utils.nameddict(
    'GroupMetadata',
    (
        'gid'
        , 'group_key'
        , 'group_key_text'
        , 'group_name'
        , 'content_type'
    ))

ContentMetadata = utils.nameddict(
    'ContentMetadata',
    (
        'cid'
        , 'content_key'
        , 'content_key_name'
#        , 'content_name'
        , 'content_type'
        , 'content_length'
        , 'last_modified'
        , 'encoding'
#        , 'md5'
        , 'order'
        , 'revert_to'
    ))

def is_compatible_with_sha1_key(key):
    try:
        return len(key) == 40 and int(key, 16) != 0
    except ValueError:
        return False

def key_for(key):
    if key:
        if is_compatible_with_sha1_key(key):
            return key
        else:
            return hashlib.sha1(key.encode('UTF-8')).hexdigest()
    else:
        raise KeyError('key is empty')

def scan_key_blobs():
    return 'BLOBS/*'

def scan_key_blobs_gid(group_id):
    return f'BLOBS/GID={group_id}*'

def incr_key_for_group_id():
    return f'BLOBS/GROUP_ID'

def zset_key_for_group_names():
    return f'BLOBS/GROUP_NAMES'

def stream_key_for_group_metadata(group_key):
    return f'BLOBS/GKEY={group_key}/METADATA'


def incr_key_for_content_id(group_id):
    return f'BLOBS/GID={group_id}/CONTENT_ID'

def zset_key_for_content_keys(group_id):
    return f'BLOBS/GID={group_id}/CONTENT_KEYS'

#def stream_key_for_contents_metadata(group_id):
#    return f'BLOBS/GID={group_id}/CONTENTS_METADATA'

#def list_key_for_versions(group_id, content_key):
#    return f'BLOBS/GID={group_id}/CKEY={content_key}/VERSIONS'

def stream_key_for_contents_metadata(group_id, content_key):
    return f'BLOBS/GID={group_id}/CKEY={content_key}/METADATA'

def obj_key_for_content(group_id, content_id):
    return f'BLOBS/GID={group_id}/CID={content_id}/OBJ'

def stream_key_for_object_buffer():
    return 'BLOBS/BUFFER'

def obj_key_for_object_buffer(session_id, buffer_id):
    return f'BLOBS/BUFFER/{session_id}/{buffer_id}'



async def get_group_metadata_with_version(redis, group_key):
    key = key_for(group_key)
    stream_key = stream_key_for_group_metadata(key)
    stream_id, kv = await redis.xlast_str_with_id(stream_key)
    if not stream_id:
        raise KeyError('group_key[{}] not found.'.format(group_key))
    return (GroupMetadata(kv), stream_id)

async def get_group_metadata(redis, group_key):
    key = key_for(group_key)
    stream_key = stream_key_for_group_metadata(key)
    kv = await redis.xlast_str(stream_key)
    if not kv:
        raise KeyError('group_key[{}] not found.'.format(group_key))
    return GroupMetadata(kv)

async def get_group_content_metadata_with_version(redis, group_key, content_key):
    group = await get_group_metadata(redis, group_key)

    key = key_for(content_key)
    stream_key = stream_key_for_contents_metadata(group.gid, key)
    stream_id, kv = await redis.xlast_str_with_id(stream_key)
    if not stream_id:
        raise KeyError('content_key [{}/{}] not found.'.format(group_key, content_key))
    return (group, ContentMetadata(kv), stream_id)
    
async def get_group_content_metadata_for(redis, group_key, content_key, version):
    group = await get_group_metadata(redis, group_key)

    key = key_for(content_key)
    stream_key = stream_key_for_contents_metadata(group.gid, key)
    ret_version, content_dict = await redis.xget_str_with_id(stream_key, version)
    content = ContentMetadata(content_dict)
    if not content or content.content_key != key:
        raise KeyError('version[{}] not found in content_key [{}/{}]'.format(version, group_key, content_key))
    return (group, content, ret_version)

def determine_last_modified(last_modified_in_request):
    if last_modified_in_request:
        try:
            dt = datetime.fromtimestamp(int(last_modified_in_request))
        except ValueError:
            dt = parsedate_to_datetime(last_modified_in_request)
    else:
       dt = datetime.now()
    return int(dt.timestamp())

async def get_content_metadata(params : dict
                               , content_key : str
#                               , content_name : str
                               , content_type : str
                               , encoding : str
                               , last_modified : str):

    metadata = ContentMetadata(params)
    if content_key:
        if is_compatible_with_sha1_key(content_key):
            metadata.content_key = content_key
#            metadata.content_name = content_name
        else:
            metadata.content_key = hashlib.sha1(content_key.encode('UTF-8')).hexdigest()
            metadata.content_key_name = content_key
#            metadata.content_name = content_name if content_name else content_key
#    metadata.content_type = content_type if content_type else guess_type(content_name)[0]
    metadata.content_type = content_type if content_type else guess_type(content_key)[0]
    metadata.last_modified = determine_last_modified(last_modified)
    return metadata




from ducts.spi import EventHandler, webapi

from io import BytesIO
from hashlib import md5
from datetime import datetime

from aiohttp import web


import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('Get Value from Redis Server')
        return handler_spec

    async def handle(self, event):
        call = self.list_versions
        if isinstance(event.data, str):
            return await call(event.data)
        elif isinstance(event.data, list):
            return await call(*event.data)
        elif isinstance(event.data, map):
            return await call(**event.data)
        elif event.data:
            raise ValueError('invalid argument. data=[{}]'.format(event.data))
        else:
            raise ValueError('group_key is required.')
        
    async def list_versions(self, group_key : str, version : str = '+', count : int = 10) -> list:
        ret = await self.manager.redis.xrevrange_str(
            self.helper.stream_key_for_group_metadata(group_key)
            , start = version, stop = '-', count = count)
        return list(ret)
        



import os
import sys
from pathlib import Path
import importlib
from importlib import machinery
import abc
import inspect

from collections import namedtuple
from collections import OrderedDict
from collections import defaultdict
from enum import Enum

import asyncio

import logging

class EventHandler(metaclass=abc.ABCMeta):

    ATTR_KEY = 'KEY'

    @abc.abstractmethod
    async def setup(self, handler_spec, manager):
        '''
        called onece when server starts setup process.
        coroutine returned by this method is awaited by the manager
        '''
        pass
    
    async def run(self, manager):
        '''
        called onece when server is started.
        coroutine returned by this method is passed to ensure_future method
        '''
        pass
    
    @abc.abstractmethod
    async def handle(self, event):
        '''
        called when messaged is passed by the client.
        '''
        pass
    
    async def handle_connected(self, event_session):
        '''
        called when the new session is started.
        '''
        pass

    async def handle_wait_reconnect(self, event_session):
        '''
        called when the connection failure is detected.
        '''
        pass

    async def handle_closed(self, event_session):
        '''
        called when the session is closed.
        '''
        pass

class HandlerSpec():

    def __init__(self, id, key, callback):
        self.id = id
        self.key = key
        self.callback = callback
        self.responsive = False
        self.description = ''
        self.webapi_context = key.lower()
        
        #self.input_sample = None
        #self.output_sample = None

    def __str__(self):
        return '{}({})-{}:{}'.format(type(self), id(self), self.id, self.key)
        
    def __repr__(self):
        return '{}({})-{}:{}'.format(type(self), id(self), self.id, self.key)
        
    def set_as_responsive(self):
        self.responsive = True
        
    def set_description(self, description):
        self.description = description

    def set_webapi_context(self, webapi_context):
        self.webapi_context = webapi_context

    def _asdict(self):
        return OrderedDict([(key,getattr(self, key))for key in ['id', 'key', 'callback', 'responsive', 'description']])
    
class webapi(object):

    apis = defaultdict(dict)

    @classmethod
    def add_route(cls, path='', method='*'):
        '''
        use this decorator to add webapi route
        '''
        def deco(func):
            cls.apis[func.__module__][path] = ('add_route', method, func.__name__)
            return func
        return deco

class EventHandlerManager(metaclass=abc.ABCMeta):

    @property
    @abc.abstractmethod
    def redis(self):
        pass

    @abc.abstractmethod
    def load_helper_module(self, module_name):
        pass

    @abc.abstractmethod
    def get_handler_module(self, event_name):
        pass

    @abc.abstractmethod
    def get_handler_for(self, event_id):
        pass

    @abc.abstractmethod
    def handlers(self):
        pass

    @abc.abstractmethod
    async def has_server_attribute(self, key):
        pass

    @abc.abstractmethod
    async def get_server_attribute(self, key):
        pass

    @abc.abstractmethod
    async def set_server_attribute(self, key, value):
        pass

    
class Event(object):

    def __init__(self, event_id, event_session, data = None):
        assert event_id >= 0, 'event_id cannot be negative.'
        assert event_session, 'event_session cannot be null.'
        self.id = event_id
        self.session = event_session
        self.data = data

        
class EventSession(metaclass=abc.ABCMeta):

    def __init__(self):
        self._local_property = {}

    @property
    def local_property(self):
        return self._local_property

    #@property
    @abc.abstractmethod
    def request_id(self):
        pass

    #@property
    @abc.abstractmethod
    def session_id(self):
        pass

    @property
    @abc.abstractmethod
    def loop(self):
        pass

    @property
    @abc.abstractmethod
    def redis(self):
        pass

    @abc.abstractmethod
    async def has_session_attribute(self, key):
        pass

    @abc.abstractmethod
    async def get_session_attribute(self, key):
        pass

    @abc.abstractmethod
    async def set_session_attribute(self, key, value):
        pass

    @abc.abstractmethod
    async def has_server_attribute(self, key):
        pass

    @abc.abstractmethod
    async def get_server_attribute(self, key):
        pass

    @abc.abstractmethod
    async def set_server_attribute(self, key, value):
        pass

    @abc.abstractmethod
    async def is_closed(self):
        pass

class RedisClient(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    async def execute(self, cmd, *args):
        pass
            
    @abc.abstractmethod
    async def execute_str(self, cmd, *args):
        pass
            
    @abc.abstractmethod
    async def subscribe(self, key):
        pass
        
    @abc.abstractmethod
    async def unsubscribe(self, key_or_channel):
        pass
        
    @abc.abstractmethod
    async def psubscribe(self, key):
        pass
        
    @abc.abstractmethod
    async def punsubscribe(self, key_or_channel):
        pass
        
    @abc.abstractmethod
    async def xadd(self, streamkey, *args, **kwargs):
        pass
    
    @abc.abstractmethod
    async def xadd_and_publish(self, pubkey, streamkey, *args, **kwargs):
        pass

    @abc.abstractmethod
    async def xget_str(self, streamkey, stream_id):
        pass

    @abc.abstractmethod
    async def xlast_str(self, streamkey):
        pass

    @abc.abstractmethod
    async def psub_and_xrange_str(self, subkey, streamkey, last_count = 0):
        pass
        
    @abc.abstractmethod
    async def psub_and_xrange_str_for_each_id(self, subkey, streamkey):
        pass




import os
import sys
from pathlib import Path
import importlib
from importlib import machinery
import abc
import inspect

from collections import namedtuple
from collections import OrderedDict
from enum import Enum

from ifconf import configure_module, config_callback

import asyncio

import logging

from .spi import HandlerSpec, EventHandler, EventHandlerManager, webapi

@config_callback
def config(loader):
    loader.add_attr_path('module_path_handler', Path('./handler'), help='module path for event_handler files')
    loader.add_attr_path('local_path_handler', Path('./handler'), help='local path for event_handler files')
    loader.add_attr('plugin_filename_pattern', 'evt_*.py', help='event handler plugin filename pattern')
    loader.add_attr_int('min_user_handler_id', 1000, help='minimum handler ID for user plugin')

    
class HandleType(Enum):
    ALIVE_MONITORING = 0
    ASYNC_FOR = 1
    FOR = 2
    ASYNC = 3
    SYNC = 4

    
class HandlerManager(EventHandlerManager):

    def __init__(self, context):
        self.logger = logging.getLogger(__name__).getChild('manager')
        self.conf = configure_module(config)
        self.context = context
        self.loop = context.loop
        self._redis = context.redis
        self._helpers = {}
        self.specs = []
        self.key_ids = {}
        self._handlers = OrderedDict()
        self._plugin_dirs = []

    @property
    def redis(self):
        return self._redis

    async def setup(self):
        for path, level in self.context.plugin_paths(self.conf.module_path_handler, self.conf.local_path_handler):
            await self.load_plugins(path, level < 2)

    def web_handlers(self):
        for t,h in self._handlers.values():
            if h.__module__ not in webapi.apis:
                self.logger.debug('WITHOUT_WEBAPI|MODULE=%s', h.__module__)
                continue
            handler_methods = {kv[0]:kv[1] for kv in inspect.getmembers(h, inspect.iscoroutinefunction)}
            context = h.SPEC.webapi_context
            context = context[1:] if context.startswith('/') else context
            context = context[0:-1] if context.endswith('/') else context
            self.logger.debug('WEBAPI|MODULE=%s|path=%s', h.__module__, webapi.apis[h.__module__].keys())
            for path, (route_add_type, method, funcname) in webapi.apis[h.__module__].items():
                path = path if path.startswith('/') else '/' + path
                path = path[0:-1] if path.endswith('/') else path
                yield (route_add_type, context + path, method, handler_methods[funcname])

    async def run(self):
        for t,h in self._handlers.values():
            f = asyncio.ensure_future(h.run(self))
            f.add_done_callback(self.done_callback)

    def done_callback(self, future):
        pass

    async def load_plugins(self, plugin_dir, is_system_handler):
        if is_system_handler:
            self.logger.debug('SYSTEM_PLUGIN_DIR=%s', plugin_dir)
        else:
            self.logger.info('PLUGIN_DIR=%s', plugin_dir)
        self._plugin_dirs.insert(0, plugin_dir)
        local_modules = set()
        for p in plugin_dir.glob(self.conf.plugin_filename_pattern):
            loader = machinery.SourceFileLoader(p.stem, str(p))
            module = loader.load_module()
            local_modules.add(module.__name__)
        for cls in sorted(EventHandler.__subclasses__(), key=lambda x: (str(x), x)):
            h = cls()
            if h.__class__.__module__ not in local_modules:
                continue
            spec = await self.setup_handler(h, is_system_handler)
            if spec is None:
                continue
            self.specs.append(spec)
            self.key_ids[spec.key] = spec.id if not spec.key in self.key_ids else max(self.key_ids[spec.key], spec.id)
            if spec.id == 0:
                htype = HandleType.ALIVE_MONITORING
            elif inspect.isasyncgenfunction(h.handle):
                htype = HandleType.ASYNC_FOR
            elif inspect.isgeneratorfunction(h.handle):
                htype = HandleType.FOR 
            elif inspect.iscoroutinefunction(h.handle):
                htype = HandleType.ASYNC
            else:
                htype = HandleType.SYNC
            self._handlers[spec.id] = (htype, h)
        if self.logger.isEnabledFor(logging.DEBUG):
            for s in self.specs:
                self.logger.debug('HANDLER|SPEC=%s', s)
        

    async def setup_handler(self, handler, is_system_handler):
        index = 0
        for ids in handler.__module__.split('_'):
            try:
                index += (len(ids) + 1)
                id = int(ids)
                setattr(handler, 'ID', id)
                key = getattr(handler, handler.ATTR_KEY, handler.__module__[index:])
                if not key:
                    key = 'undefined'
                key = key.upper()
                setattr(handler, handler.ATTR_KEY, key)
                callback = inspect.isasyncgenfunction(handler.handle) or inspect.isgeneratorfunction(handler.handle)
                #desc = handler.description()
                if id < self.conf.min_user_handler_id and not is_system_handler:
                    self.logger.warn('Permission denied. User handler ID must be less than [%s] but was [%s]', self.conf.min_user_handler_id, id)
                    return None
                spec = HandlerSpec(id, key, callback)
                ret = handler.setup(spec, self)
                setattr(handler, 'SPEC', spec)
                return ret if not inspect.isawaitable(ret) else await ret
            except ValueError:
                continue

    def load_helper_module(self, module_name):
        return self._helpers[module_name] if module_name in self._helpers else self._new_helper_module(module_name)

    def get_handler_module(self, event_name):
        return self.get_handler_for(self.key_ids[event_name])[1]
    
    def get_handler_for(self, event_id):
        return self._handlers[event_id]

    def handlers(self):
        return self._handlers.values()

    def _new_helper_module(self, module_name):
        for plugin_dir in self._plugin_dirs:
            p = plugin_dir.joinpath("{}.py".format(module_name))
            loader = machinery.SourceFileLoader(module_name, str(p))
            module = loader.load_module()
            self._helpers[module_name] = module
            return module
        
    async def get_server_attribute(self, key):
        return self.context.attrs[key]

    async def has_server_attribute(self, key):
        return key in self.context.attrs

    async def set_server_attribute(self, key, value):
        self.context.attrs[key] = value

    





from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from itertools import chain
import math
import hashlib
import struct

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_contents = KEYS[1];
local redis_key_contents_object = KEYS[2];
local redis_key_contents_keys = KEYS[3];
local content_key = ARGV[1];
local content_order = ARGV[2];
local content = ARGV[3];
if content ~= "" then
  redis.call("SET", redis_key_contents_object, content);
end
local stream_id = redis.call("XADD", redis_key_contents, "*", unpack(ARGV, 4, table.maxn(ARGV)));
redis.call("ZADD", redis_key_contents_keys, content_order, content_key);
return stream_id;
'''
    
    SCRIPT_METADATA_ONLY='''\
local redis_key_contents = KEYS[1];
local redis_key_contents_keys = KEYS[2];
local content_key = ARGV[1];
local content_order = ARGV[2];
local stream_id = redis.call("XADD", redis_key_contents, "*", unpack(ARGV, 3, table.maxn(ARGV)));
redis.call("ZADD", redis_key_contents_keys, content_order, content_key);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.update_content(event.session, event.data[0], event.data[1], event.data[2], event.data[3] if len(event.data) > 3 else {})

    async def update_content(
            self
            , session : EventSession
            , group_key : str
            , content_key : str
            , content_body : bytes
            , update_params : dict = {}):
        
        if not group_key:
            raise ValueError('group_key must be set')
        if not content_key:
            raise ValueError('content_key must be set')
        if not isinstance(update_params, dict):
            raise ValueError('update_dict must be an instance of dict')
        
        group, content, version = await self.helper.get_group_content_metadata_with_version(
            self.manager.redis, group_key, content_key)
        
        old = content.copy()
        content.update(update_params)

        if content.cid != old.cid:
            raise ValueError('cid was expeccted to be [{}] but was [{}].'.format(old.cid, content.cid))

        if content.content_key != old.content_key:
            raise ValueError('content_key was expeccted to be [{}] but was [{}].'.format(old.content_key, content.content_key))

        redis_key_contents_metadata = self.helper.stream_key_for_contents_metadata(group.gid, content.content_key)
        redis_key_contents_keys = self.helper.zset_key_for_content_keys(group.gid)
        
        if not content_body:
            if old == content:
                return version
            else:
                ret = await session.redis.evalsha(
                    Handler.SCRIPT_METADATA_ONLY
                    , 2
                    , redis_key_contents_metadata, redis_key_contents_keys
                    , content.content_key, content.order, *chain.from_iterable(content.items()))
                return {'group_key': group.group_key, 'content_key': content.content_key, 'version': ret.decode('UTF-8')}

        if isinstance(content_body, str):
            content.encoding = content.encoding if content.encoding else 'UTF-8'
            content_body = content_body.encode(content.encoding)
        elif not isinstance(content_body, bytes):
            raise ValueError('content_body must be str or bytes but was {}'.format(type(content_body)))
    
        content.content_length = len(content_body)
        content.cid = hashlib.sha1(content_body).hexdigest()
        redis_key_contents_object = self.helper.obj_key_for_content(group.gid, content.cid)
        if int(await session.redis.execute('EXISTS', redis_key_contents_object)) > 0:
            content_body = ''

        if old.last_modified == content.last_modified:
            content.last_modified = int(datetime.now().timestamp())
        else:
            content.last_modified = self.helper.determine_last_modified(content.last_modified)
        
        ret = await session.redis.evalsha(
            Handler.SCRIPT
            , 3
            , redis_key_contents_metadata, redis_key_contents_object, redis_key_contents_keys
            , content.content_key, content.order, content_body, *chain.from_iterable(content.items()))
        return {'group_key': group.group_key, 'content_key': content.content_key, 'version': ret.decode('UTF-8')}
        



from ducts.spi import EventHandler, EventSession

from io import BytesIO
from datetime import datetime
from struct import pack
from itertools import chain
import math
import hashlib

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    SCRIPT='''\
local redis_key_contents = KEYS[1];
local redis_key_contents_object = KEYS[2];
local redis_key_contents_keys = KEYS[3];
local redis_key_contents_object_buffer = KEYS[4];
local content_key = ARGV[1];
local content_order = ARGV[2];
local content_exists = ARGV[3];
if content_exists == "0" then
  redis.call("RENAME", redis_key_contents_object_buffer, redis_key_contents_object);
else
  redis.call("DEL", redis_key_contents_object_buffer);
end
local stream_id = redis.call("XADD", redis_key_contents, "*", unpack(ARGV, 4, table.maxn(ARGV)));
redis.call("ZADD", redis_key_contents_keys, content_order, content_key);
return stream_id;
'''

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')

        handler_spec.set_description('Register Resource')
        return handler_spec

    async def handle(self, event):
        return await self.add_content(event.session, event.data[0], event.data[1], **event.data[2])

    async def add_content(
            self
            , session : EventSession
            , group_key : str
            , content_buffer_key : str
            , content_key : str = ''
#            , content_name : str = ''
            , content_type : str = ''
            , encoding : str =''
            , last_modified : str = ''
            , order : int = ''
            , **other_params):
        if not group_key:
            raise ValueError('group_key must be set')
        if not content_buffer_key:
            raise ValueError('content_buffer_key must be set')
        if not isinstance(content_buffer_key, str):
            raise ValueError('content_buffer_key mus be string but was {}'.format(type(content)))

        group = await self.helper.get_group_metadata(self.manager.redis, group_key)
        metadata = await self.helper.get_content_metadata(other_params.copy()
                                                          , content_key
                                                          , content_type
                                                          , encoding
                                                          , last_modified)
        
        metadata.content_type = metadata.content_type if metadata.content_type else 'application/octet-stream'
        if encoding:
            metadata.encoding = encoding
        
        redis_key_contents_object_buffer = self.helper.obj_key_for_object_buffer(session.session_id(), content_buffer_key)
        content_length = await self.manager.redis.execute('STRLEN', redis_key_contents_object_buffer)
        sha1 = hashlib.sha1()
        buf_len = sha1.block_size * 4096
        start = 0
        while start <= content_length:
            end = start + buf_len - 1
            buf = await self.manager.redis.execute('GETRANGE', redis_key_contents_object_buffer, start, end)
            sha1.update(buf)
            start += buf_len
            
        metadata.content_length = content_length
        metadata.cid = sha1.hexdigest()

        index = await session.redis.execute('INCR', self.helper.incr_key_for_content_id(group.gid))
        metadata.order = order if order else index * 10

        redis_key_contents_object = self.helper.obj_key_for_content(group.gid, metadata.cid)
        content_exists = int(await session.redis.execute('EXISTS', redis_key_contents_object)) 
        
        if metadata.content_key:
            if int(await session.redis.execute('EXISTS', self.helper.stream_key_for_contents_metadata(group.gid, metadata.content_key))) > 0:
                raise KeyError('content_key:[{}/{}] already exits.'.format(group_key, content_key))
            else:
                pass
        else:
            metadata.content_key = hashlib.sha1(redis_key_contents_object.encode('UTF-8')).hexdigest()
        
        redis_key_contents_metadata = self.helper.stream_key_for_contents_metadata(group.gid, metadata.content_key)
        redis_key_contents_keys = self.helper.zset_key_for_content_keys(group.gid)

        ret = await session.redis.evalsha(
            Handler.SCRIPT
            , 4
            , redis_key_contents_metadata, redis_key_contents_object, redis_key_contents_keys, redis_key_contents_object_buffer
            , metadata.content_key, metadata.order, content_exists, *chain.from_iterable(metadata.items()))
        return {'group_key': group.group_key, 'content_key': metadata.content_key}
            





from ducts.spi import EventHandler

from io import BytesIO
from datetime import datetime
from hashlib import md5, sha1, sha256
from struct import pack
from itertools import chain

from mimetypes import guess_type
from email.utils import formatdate
from email.utils import parsedate_to_datetime

from aiohttp import web

import traceback
import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('[TEST_USE_ONLY] delete all blobs keys')
        return handler_spec

    async def handle(self, event):
        return await self.delete_all(event.session, event.data)

    async def delete_all(self, session, keyword):
        if keyword != "I'm crazy!":
            raise ValueError('you are too smart to delete all contents!')

        scan_key = self.helper.scan_key_blobs()
        count = 0
        async for key in session.redis.scan_for(scan_key):
            ret = await session.redis.execute('DEL', key)
            logger.info('DELETE|%s|%Ss', key, ret)
            count += 1
        return count



from ducts.spi import EventHandler, webapi

from io import BytesIO
from hashlib import md5
from datetime import datetime

from aiohttp import web


import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('Get Value from Redis Server')
        return handler_spec

    async def run(self, manager):
        self.list_groups = self.manager.get_handler_module('BLOBS_GROUP_LIST').list_groups
        self.list_contents = self.manager.get_handler_module('BLOBS_CONTENT_LIST').list_contents
        self.list_content_versions = self.manager.get_handler_module('BLOBS_CONTENT_VERSIONS').list_versions
        self.list_content_metadata = self.manager.get_handler_module('BLOBS_CONTENT_METADATA').get_metadata
        self.get_group_metadata = self.manager.get_handler_module('BLOBS_GROUP_METADATA').get_metadata
        
    async def handle(self, event):
        call = self.get_content
        if isinstance(event.data, str):
            raise ValueError('both group_key and content_key are required.')
        elif isinstance(event.data, list):
            coro = call(*event.data)
        elif isinstance(event.data, map):
            coro = call(**event.data)
        elif event.data:
            raise ValueError('invalid argument. data=[{}]'.format(event.data))
        else:
            raise ValueError('both group_key and content_key are required.')
        content_metadata, data_length = await coro.__anext__()
        async for ret in coro:
            yield ret

    async def get_content(self, group_key : str, content_key : str, version : str = '', start : int = 0, stop : int = -1):
        if version:
            group, content, version = await self.helper.get_group_content_metadata_for(
                self.manager.redis, group_key, content_key, version)
        else:
            group, content, version = await self.helper.get_group_content_metadata_with_version(
                self.manager.redis, group_key, content_key)

        redis_key_blob_data = self.helper.obj_key_for_content(group.gid,  content.cid)
        content_length = await self.manager.redis.execute('STRLEN', redis_key_blob_data)
        if content_length != int(content.content_length):
            #load from other locations
            pass

        if start == 0 and stop == -1:
            data = await self.manager.redis.execute('GET', redis_key_blob_data)
        elif start > content_length:
            raise ValueError('RequestRangeNotSatisfiable. length=[{}] but start was [{}]'.format(content_length, start))
        else:
            data = await self.manager.redis.execute('GETRANGE', redis_key_blob_data, start, stop)

        yield (content, len(data))
        
        bio = BytesIO(data)
        for buf in iter(lambda: bio.read(1024*1024), b''):
            yield buf
        yield b''
        
    @webapi.add_route(path='/groups', method='GET')
    async def group_keys(self, request):
        ret = await self.list_groups()
        return web.json_response(ret)
    
    @webapi.add_route(path='/groups/{group_key}/contents', method='GET')
    async def content_keys(self, request):
        group_key = request.match_info['group_key']
        try:
            ret = await self.list_contents(group_key)
            return web.json_response(ret)
        except KeyError as e:
            raise web.HTTPNotFound()

    @webapi.add_route(path='/groups/{group_key}/contents/{content_key}/versions', method='*')
    async def content_versions(self, request):
        group_key = request.match_info['group_key']
        content_key = request.match_info['content_key']
        try:
            ret = await self.list_content_versions(group_key, content_key)
            return web.json_response(ret)
        except KeyError as e:
            raise web.HTTPNotFound()
        
    @webapi.add_route(path='/groups/{group_key}/contents/{content_key}/metadata', method='*')
    async def content_metadata(self, request):
        group_key = request.match_info['group_key']
        content_key = request.match_info['content_key']
        try:
            ret = await self.list_content_metadata(group_key, content_key)
            return web.json_response(ret)
        except KeyError as e:
            raise web.HTTPNotFound()

    @webapi.add_route(path='/groups/{group_key}/contents/{content_key}/versions/{version}/metadata', method='*')
    async def content_metadata_of(self, request):
        group_key = request.match_info['group_key']
        content_key = request.match_info['content_key']
        version = request.match_info['version']
        try:
            ret = await self.list_content_metadata(group_key, content_key, version)
            return web.json_response(ret)
        except KeyError as e:
            raise web.HTTPNotFound()
    
    @webapi.add_route(path='/{group_key}/{content_key}/blob', method='*')
    async def service(self, request):
        logger.debug('HEADERS|%s', request.raw_headers)
        group_key = request.match_info['group_key']
        content_key = request.match_info['content_key']

        group, content, version = await self.helper.get_group_content_metadata_with_version(
            self.manager.redis, group_key, content_key)

        redis_key_blob_data = self.helper.obj_key_for_content(group.gid,  content.cid)
        content_length = await self.manager.redis.execute('STRLEN', redis_key_blob_data)
        if content_length != int(content.content_length):
            #load from other locations
            pass

        content_range = None
        if request.http_range is None or (request.http_range.start == request.http_range.stop == None):
            http_status = 200
            data = await self.manager.redis.execute('GET', redis_key_blob_data)
        elif request.http_range.start > content_length:
            err = web.HTTPRequestRangeNotSatisfiable()
            err.headers['Content-Range'] = 'bytes {}'.format(content_length)
            raise err
        else:
            http_status = 206
            start = request.http_range.start if request.http_range.start is not None else 0
            stop = request.http_range.stop -1 if request.http_range.stop is not None else content_length - 1
            data = await self.manager.redis.execute('GETRANGE', redis_key_blob_data, start, stop)
            content_range = "bytes {}-{}/{}".format(start, stop, content_length)

        data_length = len(data)
        response = web.StreamResponse(status=http_status)
        response.content_type = content.content_type
        response.content_length = data_length
        response.headers['Accept-Ranges'] = 'bytes'
        if content_range is not None:
            response.headers['Content-Range'] = content_range
        response.headers['ETag'] = content.md5
        await response.prepare(request)
        bio = BytesIO(data)
        for buf in iter(lambda: bio.read(1024*1024), b''):
            await response.write(buf)
        await response.write_eof()




    



from ducts.spi import EventHandler, webapi

from io import BytesIO
from hashlib import md5
from datetime import datetime

from aiohttp import web


import logging
logger = logging.getLogger(__name__)

class Handler(EventHandler):

    def __init__(self):
        super().__init__()

    def setup(self, handler_spec, manager):
        self.manager = manager
        self.helper = manager.load_helper_module('helper_blobs')
        handler_spec.set_description('Get Value from Redis Server')
        return handler_spec

    async def handle(self, event):
        return await self.list_groups()
    
    async def list_groups(self):
        groups = await self.manager.redis.execute_str('ZRANGEBYLEX', self.helper.zset_key_for_group_names(), '-', '+')
        return list( (vl[-1], '::'.join(vl[0:-1])) for vl in (v.split('::') for v in groups) )

    
